-----------------------------------------------------------------------------------------------
-- Client Lua Script for Radar
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"

local LibStub = _G["LibStub"]
local PKDLib = LibStub:GetLibrary("PKDLib-1.0",1)

-----------------------------------------------------------------------------------------------
-- Radar Module Definition
-----------------------------------------------------------------------------------------------
local Radar = {} 
 
local ktMarkerIsHarvest = {
	IronNode				= true,
	TitaniumNode			= true,
	ZephyriteNode			= true,
	PlatinumNode			= true,
	HydrogemNode			= true,
	XenociteNode			= true,
	ShadeslateNode			= true,
	GalactiumNode			= true,
	NovaciteNode			= true,
	StandardRelicNode		= true,
	AcceleratedRelicNode	= true,
	AdvancedRelicNode		= true,
	DynamicRelicNode		= true,
	KineticRelicNode		= true,
	SpirovineNode			= true,
	BladeleafNode			= true,
	YellowbellNode			= true,
	PummelgranateNode		= true,
	SerpentlilyNode			= true,
	GoldleafNode			= true,
	HoneywheatNode			= true,
	CrowncornNode			= true,
	CoralscaleNode			= true,
	LogicleafNode			= true,
	StoutrootNode			= true,
	GlowmelonNode			= true,
	FaerybloomNode			= true,
	WitherwoodNode			= true,
	FlamefrondNode			= true,
	GrimgourdNode			= true,
	MourningstarNode		= true,
	BloodbriarNode			= true,
	OctopodNode				= true,
	HeartichokeNode			= true,
	SmlGrowthshroomNode		= true,
	MedGrowthshroomNode		= true,
	LrgGrowthshroomNode		= true,
	SmlHarvestshroomNode	= true,
	MedHarvestshroomNode	= true,
	LrgHarvestshroomNode	= true,
	SmlRenewshroomNode		= true,
	MedRenewshroomNode		= true,
	LrgRenewshroomNode		= true,
	AlgorocTreeNode			= true,
	CelestionTreeNode		= true,
	DeraduneTreeNode		= true,
	EllevarTreeNode			= true,
	GalerasTreeNode			= true,
	AuroriaTreeNode			= true,
	WhitevaleTreeNode		= true,
	DreadmoorTreeNode		= true,
	FarsideTreeNode			= true,
	CoralusTreeNode			= true,
	MurkmireTreeNode		= true,
	WilderrunTreeNode		= true,
	MalgraveTreeNode		= true,
	HalonRingTreeNode		= true,
	GrimvaultTreeNode		= true,
	SchoolOfFishNode		= true
}

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function Radar:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here

    return o
end

function Radar:Init()
    Apollo.RegisterAddon(self,true,nil,{})
end
 

-----------------------------------------------------------------------------------------------
-- Radar OnLoad
-----------------------------------------------------------------------------------------------
function Radar:OnLoad()
    -- Register handlers for events, slash commands and timer, etc.
    -- e.g. Apollo.RegisterEventHandler("KeyDown", "OnKeyDown", self)
    Apollo.RegisterSlashCommand("radar", "OnRadarOn", self)
    Apollo.RegisterTimerHandler("Radar_Ping", "OnPing", self)

	Apollo.RegisterEventHandler("VarChange_FrameCount", "OnRadarUpdate", self)
    
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
	Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)
	Apollo.RegisterEventHandler("UnitActivationTypeChanged", "OnUnitChanged", self)
--	Apollo.RegisterEventHandler("TargetUnitChanged", "OnTargetUnitChanged", self)


    -- load our forms
    self.wndMain = Apollo.LoadForm("Radar.xml", "RadarForm", nil, self)
    self.wndMain:Show(false)

	self.wndOptions = Apollo.LoadForm("Radar.xml", "RadarOptionsForm", nil, self)
	self.wndOptions:Show(false)

	self.aTracked = {}
	self.aAll = {}
	self.Filters = {}
	self.bActive = on 
	self.bContinuousPing = false 
	self.uLastTarget = nil 
	self.nTargetEventCount = 0
	
	self.bTrackLore = true 
	self.bTrackNodes = true 
	self.nNodeTier = 0
end


-----------------------------------------------------------------------------------------------
-- Radar Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

function Radar:OnSave(eLevel) 
	if (eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character) then return end 
	
	local l,t,r,b = self.wndMain:GetAnchorOffsets()
	
	local tPos = { l = l, t = t, r = r, b = b }
	
	return {
		active = self.bActive,
		ping = self.bContinuousPing,
		windowpos = tPos, 
		lore = self.bTrackLore, 
		nodes = self.bTrackNodes, 
		nodetier = self.nNodeTier,
		filters = self.Filters 
	}
end 

function Radar:OnRestore(eLevel, tData)
	self.bActive = tData.active or false 
	self.bContinuousPing = tData.ping or false 
	self.bTrackLore = tData.lore
	self.bTrackNodes = tData.nodes
	self.nNodeTier = tData.nodetier or 0 
	self.Filters = tData.filters or {} 
	
	if (tData.windowpos) then 
		local tL = tData.windowpos
		self.wndMain:SetAnchorOffsets(tL.l, tL.t, tL.r, tL.b)
	end 
	
	self.wndMain:FindChild("PingButton"):SetCheck(self.bContinuousPing)
	
	self:RebuildTracked()
end 

function Radar:OnConfigure()
	self:OnOptionsButton()
end 

-- on SlashCommand "/radar"
function Radar:OnRadarOn(strCommand, strParam)
	if (strParam == "on") then 
		self.bActive = true 
		self:RebuildList()
		if (self.bContinuousPing) then Apollo.CreateTimer("Radar_Ping",5,true) end 
		if (table.getn(self.aTracked) ~= 0) then 
			self.wndMain:Show(true) 
		end 
	elseif (strParam == "off") then 
		self.bActive = false 
		Apollo.StopTimer("Radar_Ping")
		self.wndMain:Show(false)		
	elseif (strParam == "ping") then 
		self:OnPing()
	elseif (strParam == "options") then 
		self:OnOptionsButton()
	elseif string.find(string.lower(strParam), "filter") ~= nil then
		-- Contributed by Scelestic; name-based matching for radar.
		strParam = string.lower(strParam)
		local args = {}
		for word in string.gmatch(strParam, "%a+") do table.insert(args, word) end
		if args[2] == "dump" then
			ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Command,"Current Radar Filters:")
			for word, _ in pairs(self.Filters) do 
				ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Command,"  " .. word) 
			end
		elseif args[2] == "add" then
			for i=3, #args do
				if not self.Filters[args[i]] then
					ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Command,"Added " .. args[i] .. " to Radar filters")
					self.Filters[args[i]] = true
					self:RebuildTracked()
				end
			end
		elseif args[2] == "remove" then
			for i=3, #args do
				ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Command,"Removed " .. args[i] .. " from Radar filters")
				self.Filters[args[i]] = nil
				self:RebuildTracked()
			end
		end
		
	else
		Print("\n/radar on - Enable Radar window\n/radar off - Disable Radar window\n/radar ping - 'Ping' all items Radar is currently aware of (whether or not window is enabled)\n/radar filter <dump/add/remove> [word to watch for in unit names]")
	end 
end

-- on timer
function Radar:OnPing()
	for _, tRecord in ipairs(self.aTracked) do 
		tRecord.unit:ShowHintArrow()
	end 
end

function Radar:AddToGrid(unit, nDistance) 
	if (unit == nil) then return end 

	local wndGrid = self.wndMain:FindChild("RadarGrid")
	
	local iCurrRow = wndGrid:AddRow("")
	wndGrid:SetCellLuaData(iCurrRow, 1, unit)
	wndGrid:SetCellText(iCurrRow, 1, unit:GetName())
	wndGrid:SetCellText(iCurrRow, 2, string.format("%.2f m", nDistance))
end 

function Radar:DistanceToUnit(unit)
	local uPlayer = GameLib.GetPlayerUnit()
	
	if (uPlayer == nil) then return 0 end 
 
	local posPlayer = uPlayer:GetPosition()  
	local posTarget = unit:GetPosition()
	local nDeltaX = posTarget.x - posPlayer.x
	local nDeltaY = posTarget.y - posPlayer.y
	local nDeltaZ = posTarget.z - posPlayer.z

	return math.sqrt(math.pow(nDeltaX, 2) + math.pow(nDeltaY, 2) + math.pow(nDeltaZ, 2))
end 

function Radar:PositionsEqual(tPos1, tPos2)
	return (tPos1.x == tPos2.x) and (tPos1.y == tPos2.y) and (tPos1.z == tPos2.z) 
end 

function Radar:OnRadarUpdate()
	if not self.bActive then return end 
	if not self.wndMain:IsShown() then return end 

	local player = GameLib.GetPlayerUnit()
	if (player == nil) then return end 
	
	local tPlayerPosition = player:GetPosition()
	if self:PositionsEqual(tPlayerPosition,self.tLastPlayerPos or { x = 0, y = 0, z = 0}) then return end 
	self.tLastPlayerPos = tPlayerPosition 
	
	for idx,tRecord in ipairs(self.aTracked) do 
		tRecord.distance = self:DistanceToUnit(tRecord.unit)
	end 
	
	table.sort(self.aTracked,function(a,b) if (a.distance == b.distance) then return a.unit:GetName() < b.unit:GetName() else return a.distance < b.distance end end) 

	self:RebuildList()	
end 

function Radar:RemoveRow(nRow) 
	local wndGrid = self.wndMain:FindChild("RadarGrid")

	wndGrid:DeleteRow(nRow) 
end 

function Radar:RebuildTracked()
	self.aTracked = {}
	
	for idx, drUnit in ipairs(self.aAll) do 
		if self:UnitMatchesFilters(drUnit) then
			self:AddUnit(drUnit)
		end
	end
	
	self:RebuildList()
	
	if (table.getn(self.aTracked) == 0) then self.wndMain:Show(false) end
end

function Radar:RebuildList()
	local wndGrid = self.wndMain:FindChild("RadarGrid")
	
	wndGrid:DeleteAll()
	
	for idx, tRecord in ipairs(self.aTracked) do 
		self:AddToGrid(tRecord.unit, tRecord.distance) 
	end 
end 

function Radar:AddUnit(unit) 
	if (self:RowForUnit(unit) == 0) then 
		nDistance = self:DistanceToUnit(unit)
		table.insert(self.aTracked,{ unit = unit, distance = nDistance }) 
		self:AddToGrid(unit,nDistance) 
		
		if (self.bActive) then self.wndMain:Show(true) end 
	end 
end 

function Radar:RemoveUnit(unit) 
	local nRow = self:RowForUnit(unit) 
	
	if (nRow ~= 0) then 
		table.remove(self.aTracked,nRow) 
		self:RemoveRow(nRow)
		
		if (table.getn(self.aTracked) == 0) then self.wndMain:Show(false) end 
	end 
end 

function Radar:ForgetUnit(unit)
	for idx, drRecord in ipairs(self.aAll) do 
		if drRecord == unit then 
			table.remove(self.aAll,idx) 
			return
		end 
	end 
	
	return
end

function Radar:RowForUnit(unit) 
	for idx, tRecord in ipairs(self.aTracked) do 
		if tRecord.unit == unit then return idx end 
	end 
	
	return 0
end 

function Radar:KnownUnit(unit) 
	for idx, drRecord in ipairs(self.aAll) do 
		if drRecord == unit then return true end 
	end 
	
	return false
end 

function Radar:UnitMatchesFilters(unit)
	if unit == nil then return false end

	local tActivation = unit:GetActivationState()
	local strMarker = unit:GetMiniMapMarker()
	
	if (self.bTrackLore and tActivation ~= nil and tActivation.Datacube ~= nil) then 
		return true
	elseif (self.bTrackNodes and ktMarkerIsHarvest[strMarker]) then 
		return true
	else
		for n,_ in pairs(self.Filters) do
			if string.find(string.lower(unit:GetName()), n) ~= nil then
				return true
			end
		end
	end 
	
	return false
end

function Radar:OnUnitCreated(unit) 
	table.insert(self.aAll,unit)
	if self:UnitMatchesFilters(unit) then
		self:AddUnit(unit)
	end
end 

function Radar:OnUnitDestroyed(unit) 
	self:RemoveUnit(unit)
	self:ForgetUnit(unit)
end 

function Radar:OnUnitChanged(unit) 
	local bMatches = self:UnitMatchesFilters(unit)

	if (self:RowForUnit(unit) == 0) and bMatches then 
		self:AddUnit(unit)
	elseif (self:RowForUnit(unit) ~= 0) and not bMatches then 
		self:RemoveUnit(unit)
	end
end 

-----------------------------------------------------------------------------------------------
-- RadarForm Functions
-----------------------------------------------------------------------------------------------
function Radar:OnPingToggle( wndHandler, wndControl, eMouseButton )
	self.bContinuousPing = self.wndMain:FindChild("PingButton"):IsChecked()
	
	if (self.bContinuousPing) then 
		self:OnPing()
		Apollo.CreateTimer("Radar_Ping",5,true)
	else 
		Apollo.StopTimer("Radar_Ping")
	end
end

function Radar:OnRadarItemPick(wndControl, wndHandler, iRow, iCol)
	local tRecord = self.aTracked[iRow] 
	
	tRecord.unit:ShowHintArrow()	
end 

function Radar:OnOptionsButton( wndHandler, wndControl, eMouseButton )
	
	self.wndOptions:FindChild("LoreToggleBtn"):SetCheck(self.bTrackLore) 
	self.wndOptions:FindChild("NodesToggleBtn"):SetCheck(self.bTrackNodes) 
	
	self.wndOptions:Show(true) 

end

---------------------------------------------------------------------------------------------------
-- RadarOptionsForm Functions
---------------------------------------------------------------------------------------------------
function Radar:OnOptionsCancel( wndHandler, wndControl, eMouseButton )
	self.wndOptions:Show(false) 
end

function Radar:OnOptionsAccept( wndHandler, wndControl, eMouseButton )
	self.bTrackLore = self.wndOptions:FindChild("LoreToggleBtn"):IsChecked()
	self.bTrackNodes = self.wndOptions:FindChild("NodesToggleBtn"):IsChecked()
	
	self.wndOptions:Show(false) 
	self:RebuildTracked()
end

-----------------------------------------------------------------------------------------------
-- Radar Instance
-----------------------------------------------------------------------------------------------
local RadarInst = Radar:new()
RadarInst:Init()
